#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NKEYS 12
#define STRMAX 1024

/* ========================================================================== *
 * shiftxy
 *
 * Программа читает объекты из входного файла MIF,
 * сдвигает их на (dx, dy) в единицах XY units слоя,
 * записывает в выходной файл MIF.
 *
 * Usage: shiftxy <input> <output> <dx> <dy>
 *
 * Аргументы:
 *     input  - входной файл MIF
 *     output - выходной файл MIF
 *     dx     - сдвиг по координате X
 *     dy     - сдвиг по координате Y
 *
 * Границы слоя не изменяются.
 *
 * Действий с файлом MID не производится,
 * просто используйте оригинал после копирования или перименования.
 * -------------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
  char buf[STRMAX], w[8][128];
  char *keyword[NKEYS] = {
    "Arc", "Ellipse", "Line", "Pline", "Point", "Region", "Rect",
    "Roundrect", "Text", "Multipoint", "Center", "Multiple"
  };
  double dx, dy, x, y, x2, y2;
  int nsec, npt, t, k;
  FILE *fp0, *fp1;

  if (argc < 5) {
    puts("usage: shiftxy <input> <output> <dx> <dy>");
    exit(EXIT_SUCCESS);
  }
  if ((fp0 = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "can't open %s\n", argv[1]);
    exit(EXIT_FAILURE);
  }
  if ((fp1 = fopen(argv[2], "w")) == NULL) {
    fprintf(stderr, "can't create %s\n", argv[2]);
    exit(EXIT_FAILURE);
  }
  dx = atof(argv[3]);
  dy = atof(argv[4]);
  nsec = npt = t = 0;
  while (fgets(buf, STRMAX, fp0) != NULL) {
    if (npt == 0) {
      if (nsec == 0) {
	switch (t) {
	case 2: /* Text string */
	  fputs(buf, fp1);
	  t = 1;
	  break;
	case 1: /* Text bbox */
	  sscanf(buf, "%s %s %s %s", w[0], w[1], w[2], w[3]);
	  x = atof(w[0]) + dx;
	  y = atof(w[1]) + dy;
	  x2 = atof(w[2]) + dx;
	  y2 = atof(w[3]) + dy;
	  fprintf(fp1, "    %f %f %f %f\n", x, y, x2, y2);
	  t = 0;
	  break;
	default:
	  sscanf(buf, "%s %s %s %s %s %s %s",
		 w[0], w[1], w[2], w[3], w[4], w[5], w[6]);
	  for (k = 0; k < NKEYS - 1; k++) {
	    if (strcmp(w[0], keyword[k]) == 0)
	      break;
	  }
	  switch (k) {
	  case 8: /* "Text" */
	    t = 2;
	    fputs(buf, fp1);
	    break;
	  case 9: /* "Multipoint" */
	    npt = atoi(w[1]);
	    fputs(buf, fp1);
	    break;
	  case 5: /* "Region" */
	    nsec = atoi(w[1]);
	    fputs(buf, fp1);
	    break;
	  case 3: /* "Pline" */
	    if (strcmp(w[1], keyword[NKEYS - 1]) == 0) /* "Pline Multiple" */
	      nsec = atoi(w[2]);
	    else
	      npt = atoi(w[1]);
	    fputs(buf, fp1);
	    break;
	  case 4: /* "Point" */
	    x = atof(w[1]) + dx;
	    y = atof(w[2]) + dy;
	    fprintf(fp1, "%s %f %f\n", w[0], x, y);
	    break;
	  case 10: /* "Center" */
	    x = atof(w[1]) + dx;
	    y = atof(w[2]) + dy;
	    fprintf(fp1, "    %s %f %f\n", w[0], x, y);
	    break;
	  case 1: /* "Ellipse" */
	  case 2: /* "Line" */
	  case 6: /* "Rect" */
	    x = atof(w[1]) + dx;
	    y = atof(w[2]) + dy;
	    x2 = atof(w[3]) + dx;
	    y2 = atof(w[4]) + dy;
	    fprintf(fp1, "%s %f %f %f %f\n",
		    w[0], x, y, x2, y2);
	    break;
	  case 7: /* "Roundrect" */
	    x = atof(w[1]) + dx;
	    y = atof(w[2]) + dy;
	    x2 = atof(w[3]) + dx;
	    y2 = atof(w[4]) + dy;
	    fprintf(fp1, "%s %f %f %f %f %s\n",
		    w[0], x, y, x2, y2, w[5]);
	    break;
	  case 0: /* "Arc" */
	    x = atof(w[1]) + dx;
	    y = atof(w[2]) + dy;
	    x2 = atof(w[3]) + dx;
	    y2 = atof(w[4]) + dy;
	    fprintf(fp1, "%s %f %f %f %f %s %s\n",
		    w[0], x, y, x2, y2, w[5], w[6]);
	    break;
	  default:
	    fputs(buf, fp1);
	  }
	}
      } else {
	npt = atoi(buf);
	fputs(buf, fp1);
	nsec--;
      }
    } else {
      sscanf(buf, "%s %s", w[0], w[1]);
      x = atof(w[0]) + dx;
      y = atof(w[1]) + dy;
      fprintf(fp1, "%f %f\n", x, y);
      npt--;
    }
  }
  fclose(fp1);
  fclose(fp0);
  return 0;
}
/* ========================================================================== */
